#ifndef UNDER_CE
#include <stdio.h>
#include <errno.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "stack.h"
#include "buildin_function.h"
#include "machine.h"


static myfloat degree_to_radian(myfloat deg);
static myfloat radian_to_degree(myfloat rad);
static myfloat my_fact(myfloat x);
static myfloat my_log(myfloat b, myfloat x);

#ifdef USE_LONG_DOUBLE

#define PI 3.141592653589793238462643383279502884L

#include "quadmath/fp.h"
static struct buildin_function_s const buildin_function[]=
{
	{"abs", 	1, fabsl},
	{"d2r", 	1, degree_to_radian},
	{"r2d", 	1, radian_to_degree},
	{"exp", 	1, expl},
	{"ln",  	1, logl},
	{"sin", 	1, sinl},
	{"cos", 	1, cosl},
	{"tan", 	1, tanl},
	{"asinh", 	1, asinhl},
	{"acosh", 	1, acoshl},
	{"atanh", 	1, atanhl},
	{"sqrt", 	1, sqrtl},
	{"fact", 	1, my_fact},
	{"atan", 	1, atanl},
	{"log10", 	1, log10l},
	{"log", 	2, my_log},
	{"asin", 	1, asinl},
	{"acos", 	1, acosl},
	{"sinh", 	1, sinhl},
	{"cosh", 	1, coshl},
	{"tanh", 	1, tanhl},
};
/* base b  ΐ֐ */
static myfloat my_log(myfloat b, myfloat x)
{
	return logl(x)/logl(b);
}


#else /* !USE_LONG_DOUBLE */

static myfloat my_asinh(myfloat x);
static myfloat my_acosh(myfloat x);
static myfloat my_atanh(myfloat x);


#ifdef RUPUTER
	static myfloat my_asin(myfloat x);
	static myfloat my_acos(myfloat x);
	static myfloat my_sinh(myfloat x);
	static myfloat my_cosh(myfloat x);
	static myfloat my_tanh(myfloat x);
	static myfloat my_log10(myfloat x);
#endif

static struct buildin_function_s const buildin_function[]=
{
	{"abs", 	1,	fabs},
	{"d2r", 	1,	degree_to_radian},
	{"r2d", 	1,	radian_to_degree},
#ifndef DEBUG
	{"exp", 	1,	exp},
	{"ln", 		1,	log},
	{"sin", 	1,	sin},
	{"cos", 	1,	cos},
	{"tan", 	1,	tan},
	{"asinh", 	1,	my_asinh},
	{"acosh", 	1,	my_acosh},
	{"atanh", 	1,	my_atanh},
	{"sqrt", 	1,	sqrt},
	{"fact", 	1,	my_fact},
	{"atan", 	1,	atan},
	{"log", 	2, my_log},
#  ifdef RUPUTER
		{"log10", 	1,	my_log10},
		{"asin", 	1,	my_asin},
		{"acos", 	1,	my_acos},
		{"sinh", 	1,	my_sinh},
		{"cosh", 	1,	my_cosh},
		{"tanh", 	1,	my_tanh},
#  else
		{"log10", 	1,	log10},
		{"asin", 	1,	asin},
		{"acos", 	1,	acos},
		{"sinh", 	1,	sinh},
		{"cosh", 	1,	cosh},
		{"tanh", 	1,	tanh},
#  endif /* RUPUTER */
#endif /* DEBUG */
};


/**************************** ȉA֐ *****************************/

#ifdef DOUBLE_IS_32BITS
 #define EPS5 0.04  /* FLT_EPSILON  1/5 x */
#else
 #define EPS5 7e-4  /* DBL_EPSILON  1/5 x */
#endif

#define LN10 2.30258509299404568401799145468436
#define LN2 0.693147180559945309417232121458177
#define PI 3.1415926535897932384626433832795

#ifndef DEBUG

static myfloat my_log(myfloat b, myfloat x)
{
	return log(x)/log(b);
}

static myfloat my_asinh(myfloat x)
{
	if (x > EPS5)
		return log(sqrt(x*x+1)+x);
	else if (x < -EPS5)
		return -log(sqrt(x*x+1)-x);
	else
		return x*(1-x*x/6);	/* eC[WJ x 0 ɏ\߂*/
}

/* ֐΂킩ƂAx >=1 łȂΒ`Ȃ */
static myfloat my_acosh(myfloat x)
{
	return log(x+sqrt(x*x-1));
}

static myfloat my_atanh(myfloat x)
{
	if (fabs(x) > EPS5)
		return 0.5*log((1+x)/(1-x));
	else
		return x*(1+x*x/3.0);	/* eC[WJ */
}

#ifdef RUPUTER
/* tOp֐ */
static myfloat my_asin(myfloat x)
{
	return atan(x/sqrt(1-x*x));
}
static myfloat my_acos(myfloat x)
{
	if(x == -1)
		x=PI;
	else
	{
		x=atan(sqrt(1-x*x)/x);
		if(x < 0)
			x+=PI;
	}
	return x;
}

static myfloat my_sinh(myfloat x)
{
	if (fabs(x) > EPS5)
	{
		x=exp(x);
		x=(x-1/x)/2;
	}
	else
		x=x*(1+x*x/6);
	return x;
}
static myfloat my_cosh(myfloat x)
{
	x=exp(x);
	x=(x+1/x)/2;
	return x;
}
static myfloat my_tanh(myfloat x)
{
	if(x > EPS5)
		x=2/(1+exp(-2*x))-1;
	else if(x < -EPS5)
		x=1-2/(exp(2*x)+1);
	else
		x=x*(1-x*x/3);
	return x;
}
static myfloat my_log10(myfloat x)
{
	return log(x)/LN10;
}

#endif /* RUPUTER */




#endif /* DEBUG */


#endif /* USE_LONG_DOUBLE */



/* ́cΕ */
static myfloat degree_to_radian(myfloat deg)
{
	return deg*PI/180;
}
static myfloat radian_to_degree(myfloat rad)
{
	return rad*180/PI;
}

/* ǨvZ */
static myfloat my_fact(myfloat x)
{
	unsigned long n;
	myfloat res;

	if(x < 0.0)
	{
		errno=EDOM;
		return 0.0;
	}

	n=(unsigned long)x;
	if(fabsl(x-n) > 0)
	{
		errno=EDOM;
		return 0.0;
	}

	res=1;
	while(n)
		res*=(myfloat)n--;

	return res;
}


struct buildin_function_s *check_build_in_function(char * const name)
{
	int i;

	for(i=0; i < sizeof(buildin_function)/sizeof(struct buildin_function_s); i++)
		if(strcmp(buildin_function[i].name, name) == 0)
			return buildin_function+i;

	return NULL;

}


