#ifndef UNDER_CE
#include <stdio.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifndef UNDER_CE
#include <errno.h>
#endif

#include "code.h"
#include "parse.h"
#include "lex.h"
#include "stack.h"
#include "variable.h"
#include "function.h"
#include "user_func.h"
#include "../libmisc/misc.h"
#include "utility.h"
#include "machine.h"

#ifdef USE_LONG_DOUBLE
long double powl(long double x, long double y);
#endif

static void do_code_comma(void);
static void do_code_subst(struct lex_s * const p);
static void do_code_inv(void);
static int do_code_not(void);
static int do_code_binop(int op);
static int do_code_binop_sub(int op, struct stack_s * const s1, struct stack_s * const s2);

void print_code(void);

int code_count;
struct lex_s *code;

static int max_code_count;

extern int def_function;

#define CODE_FIRST 32
#define CODE_STEP 8

void init_code(void)
{
	if(max_code_count != CODE_FIRST)
	{
		code=my_realloc(code, CODE_FIRST*sizeof(struct lex_s));
		max_code_count=CODE_FIRST;
	}

	code_count=0;
}

void add_code(int op, struct lex_s * const lx)
{
	if(code_count >= max_code_count)
	{
		max_code_count+=CODE_STEP;
		code=my_realloc(code, max_code_count*sizeof(struct lex_s));
	}

	if(lx)
		code[code_count]=*lx;

	code[code_count].type=op;
	code_count++;
}

void free_code(void)
{
	int i;

	for(i=0; i < code_count; i++)
	{
		switch(code[i].type)
		{
		case TYPE_UNDEF_VARIABLE:
		case TYPE_VARIABLE:
		case TYPE_FUNCTION:
		case OP_SUBST:
			free(code[i].data.id);
		}
	}
}

char *run_code(void)
{
	int i;
	struct lex_s *p;

#ifdef DEBUG_OUTPUT
	print_code();
#endif

	if(def_function == 1)
	{
		return func_def_user_function();
	}

	p=code;
	for(i=0; i < code_count && error == 0; i++, p++)
	{
		switch(p->type)
		{
		case TYPE_FLOAT:
			push_float(p->data.num.f);
			break;
		case TYPE_INTEGER:
			push_integer(p->data.num.integer);
			break;
		case TYPE_VARIABLE:
			push_num(p->variable->type, p->variable->num);
			break;
		case TYPE_UNDEF_VARIABLE:
			error=ER_UNDEF_VARIABLE;
			return p->aux;
		case TYPE_FUNCTION:
			if(call_function(p->data.id) == 1)
			{
				if(error == 0)
					error=ER_UNDEF_FUNCTION;
				return p->aux;
			}
			break;
		case TYPE_PARAMETER:
			push_parameter(p->data.num.integer.i);
			break;
		case OP_INV:
			do_code_inv();
			break;
		case OP_NOT:
			if(do_code_not() == 1)
				return p->aux;
			break;
		case OP_SUBST:
			do_code_subst(p);
			break;
		case OP_COMMA:
			do_code_comma();
			break;
		default:
			if(do_code_binop(p->type) == 1)	/* error */
				return p->aux;
		}
	}
	return NULL;
}


static void do_code_comma(void)
{
	struct stack_s *s;

	s=pop();
	(void)pop();

	push_stack(s);
}


static int do_code_not(void)
{
	struct stack_s *s;

	s=virtual_pop();

	if(s->type == TYPE_FLOAT)
	{
		error=ER_UNACCEPTED_OPRATION;
		return 1;
	}

	s->num.integer.i=~s->num.integer.i;

	return 0;
}
static void do_code_subst(struct lex_s * const p)
{
	struct stack_s *s;

	s=virtual_pop();

	if(p->variable == NULL)
	{
		variable_set(p->data.id, s);
/* ϐo^ŁAid ̃|C^tbN̂ŎgȂ */
		p->data.id=NULL;
	}
	else
	{
		p->variable->type=s->type;
		p->variable->num=s->num;
	}
}

static void do_code_inv(void)
{
	struct stack_s *s;

	s=virtual_pop();

	if(s->type == TYPE_INTEGER)
		s->num.integer.sign^=1;
	else if(s->type == TYPE_FLOAT)
		s->num.f=-s->num.f;
}

/*
 * ܂OƂāǍ^ɂ͓BЂƂ myfloat ŁA 
 * long B
 * lZA]AwvZíAd myfloat, i  long ̈Ӂj
 *   s1 s2
 *   d  d => d
 *   d  i => d
 *   i  i => i or I[o[t[ d 
 *
 * rbgZi_ρA_aArI_aAVtgj
 *   s1 s2
 *   d  d => error
 *   d  i => error
 *   i  i => i
 *
 */
static int do_code_binop(int op)
{
	struct stack_s *s1, *s2;

	s2=pop();
	s1=virtual_pop();
/*	s1 @ s2 */

	return do_code_binop_sub(op, s1, s2);
}


/* s1 <- s1 @ s2 */
static int do_code_binop_sub(int op, struct stack_s * const s1, struct stack_s * const s2)
{
	int t;

	t=(s1->type == TYPE_INTEGER ? 1 : 0);
	if(s2->type == TYPE_INTEGER)
		t|=2;

	switch(t)
	{
	case 0:	/* d @ d */
		switch(op)
		{
		case OP_ADD:
			s1->num.f+=s2->num.f;
			break;
		case OP_SUB:
			s1->num.f-=s2->num.f;
			break;
		case OP_MUL:
			s1->num.f*=s2->num.f;
			break;
		case OP_DIV:
			if(s2->num.f == 0.0)
			{
				error=ER_DIVIDE_BY_ZERO;
				return 1;
			}
			s1->num.f/=s2->num.f;
			break;
		case OP_MOD:
			if(s2->num.f == 0.0)
			{
				error=ER_MODULO_BY_ZERO;
				return 1;
			}
			s1->num.f=fmod(s1->num.f, s2->num.f);
			break;
		case OP_POW:
			errno=0;

#ifdef USE_LONG_DOUBLE
			s1->num.f=powl(s1->num.f, s2->num.f);
#else
			s1->num.f=pow(s1->num.f, s2->num.f);
#endif
			if(errno == EDOM)
			{
				error=ER_DOMAIN;
				return 1;
			}
			break;
		case OP_OR:
		case OP_AND:
		case OP_XOR:
		case OP_LSHIFT:
		case OP_RSHIFT:
			error=ER_UNACCEPTED_OPRATION;
			return 1;
		}
		break;
	case 1:	/* i @ d ok */
		switch(op)
		{
		case OP_ADD:
			s1->type=TYPE_FLOAT;
			if(s1->num.integer.sign == PLUS)
				s1->num.f=s1->num.integer.i+s2->num.f;
			else
				s1->num.f=-(myfloat)s1->num.integer.i+s2->num.f;
			break;
		case OP_SUB:
			s1->type=TYPE_FLOAT;
			if(s1->num.integer.sign == PLUS)
				s1->num.f=s1->num.integer.i-s2->num.f;
			else
				s1->num.f=-(myfloat)s1->num.integer.i-s2->num.f;
			break;
		case OP_MUL:
			s1->type=TYPE_FLOAT;

/* vZゾƁAsign ςĂ܂ip̂̂߁j
 * ̂߁As2 ̕̕𔽓]Ăidiv lj
 */
			if(s1->num.integer.sign == MINUS)
				s2->num.f=-s2->num.f;
			s1->num.f=s1->num.integer.i*s2->num.f;
			break;
		case OP_DIV:
			if(s2->num.f == 0.0)
			{
				error=ER_DIVIDE_BY_ZERO;
				return 1;
			}

			s1->type=TYPE_FLOAT;
			if(s1->num.integer.sign == MINUS)
				s2->num.f=-s2->num.f;
			s1->num.f=s1->num.integer.i/s2->num.f;
			break;
		case OP_MOD:
		{
			if(s2->num.f == 0.0)
			{
				error=ER_MODULO_BY_ZERO;
				return 1;
			}
			conv_integer_to_float(s1);

			s1->num.f=fmod(s1->num.f, s2->num.f);
			break;
		}
		case OP_POW:
		{
			conv_integer_to_float(s1);

			errno=0;
#ifdef USE_LONG_DOUBLE
			s1->num.f=powl(s1->num.f, s2->num.f);
#else
			s1->num.f=pow(s1->num.f, s2->num.f);
#endif
			if(errno == EDOM)
			{
				error=ER_DOMAIN;
				return 1;
			}
			break;
		}
		case OP_OR:
		case OP_AND:
		case OP_XOR:
		case OP_LSHIFT:
		case OP_RSHIFT:
			error=ER_UNACCEPTED_OPRATION_RIGHT;
			return 1;
		}
		break;
	case 2:	/* d @ i ok */
		switch(op)
		{
		case OP_SUB:
			s2->num.integer.sign^=1;
		case OP_ADD:
			s1->type=TYPE_FLOAT;
			if(s2->num.integer.sign == PLUS)
				s1->num.f+=s2->num.integer.i;
			else
				s1->num.f-=s2->num.integer.i;
			break;
		case OP_MUL:
			s1->type=TYPE_FLOAT;
			if(s2->num.integer.sign == MINUS)
				s1->num.f=-s1->num.f;
			s1->num.f*=s2->num.integer.i;
			break;
		case OP_DIV:
			if(s2->num.integer.i == 0)
			{
				error=ER_DIVIDE_BY_ZERO;
				return 1;
			}

			s1->type=TYPE_FLOAT;
			if(s2->num.integer.sign == MINUS)
				s1->num.f=-s1->num.f;
			s1->num.f/=s2->num.integer.i;
			break;
		case OP_MOD:
		{
			if(s2->num.integer.i == 0)
			{
				error=ER_MODULO_BY_ZERO;
				return 1;
			}

			conv_integer_to_float(s2);
			s1->num.f=fmod(s2->num.f, s2->num.f);
			break;
		}
		case OP_POW:
		{
			conv_integer_to_float(s2);

			errno=0;
#ifdef USE_LONG_DOUBLE
			s1->num.f=powl(s1->num.f, s2->num.f);
#else
			s1->num.f=pow(s1->num.f, s2->num.f);
#endif
			if(errno == EDOM)
			{
				error=ER_DOMAIN;
				return 1;
			}
			break;
		}
		case OP_OR:
		case OP_AND:
		case OP_XOR:
		case OP_LSHIFT:
		case OP_RSHIFT:
			error=ER_UNACCEPTED_OPRATION_LEFT;
			return 1;
		}

		break;
	case 3:	/* i @ i  ok*/
		switch(op)
		{
		case OP_SUB:
			s2->num.integer.sign^=1;
		case OP_ADD:
		{
			if(s1->num.integer.sign == s2->num.integer.sign)
			{
				unsigned long max=0xffffffffUL; /* max */

				max-=s1->num.integer.i;
				if(max < s2->num.integer.i)
				{
/* overflow */
					s1->type=TYPE_FLOAT;
					s1->num.f=(myfloat)s1->num.integer.i+s2->num.integer.i;
					if(s2->num.integer.sign == MINUS)
						s1->num.f=-s1->num.f;
				}
				else
					s1->num.integer.i=s1->num.integer.i+s2->num.integer.i;

				break;
			}
			else if(s1->num.integer.i < s2->num.integer.i)
			{
/* s1 ̂ŁAs2 ̕ɂȂ */
				s1->num.integer.i=s2->num.integer.i-s1->num.integer.i;
				s1->num.integer.sign=s2->num.integer.sign;
			}
			else if(s1->num.integer.i > s2->num.integer.i)
				s1->num.integer.i-=s2->num.integer.i;
			else
				s1->num.integer.i=0;
/* s1 ̕傫̂ŁAs1 ̕  */

			break;
		}
		case OP_MUL:
		{
			if(s1->num.integer.i != 0)
			{
				unsigned long max=0xffffffffUL;	/* max value */

				max/=s1->num.integer.i;

/* ̎_ŁAmax zl s1 ɂƃI[o[t[ɂȂ邱Ƃ */
				if(max < s2->num.integer.i)
				{
/* overflow */
					int flag=0;
					if(s1->num.integer.sign != s2->num.integer.sign)
						flag=1;

					s1->type=TYPE_FLOAT;
					s1->num.f=(myfloat)s1->num.integer.i*s2->num.integer.i;

					if(flag)
						s1->num.f=-s1->num.f;
				}
				else
				{
					s1->num.integer.i*=s2->num.integer.i;
					s1->num.integer.sign^=s2->num.integer.sign;
				}
			}
			break;
		}
		case OP_DIV:
		{
			if(s2->num.integer.i == 0)
			{
/* divide by zero */
				error=ER_DIVIDE_BY_ZERO;
				return 1;
			}
			else
			{
				if(s1->num.integer.i%s2->num.integer.i == 0)
				{
	/* ؂ */
					s1->num.integer.i/=s2->num.integer.i;
					s1->num.integer.sign^=s2->num.integer.sign;
				}
				else
				{
	/* underflow */
					int flag=0;

					if(s1->num.integer.sign != s2->num.integer.sign)
						flag=1;

					s1->type=TYPE_FLOAT;
					s1->num.f=(myfloat)s1->num.integer.i/s2->num.integer.i;
					if(flag)
						s1->num.f=-s1->num.f;
				}
			}
			break;

		}
		case OP_MOD:
			if(s2->num.integer.i == 0)
			{
/* divide by zero */
				error=ER_MODULO_BY_ZERO;
				return 1;
			}
			else
				s1->num.integer.i%=s2->num.integer.i;
			break;
		case OP_POW:
		{
			conv_integer_to_float(s1);
			conv_integer_to_float(s2);

			errno=0;
#ifdef USE_LONG_DOUBLE
			s1->num.f=powl(s1->num.f, s2->num.f);
#else
			s1->num.f=pow(s1->num.f, s2->num.f);
#endif
			if(errno == EDOM)
			{
				error=ER_DOMAIN;
				return 1;
			}

			break;
		}
		case OP_OR:
			s1->num.integer.i|=s2->num.integer.i;
			break;
		case OP_AND:
			s1->num.integer.i&=s2->num.integer.i;
			break;
		case OP_XOR:
			s1->num.integer.i^=s2->num.integer.i;
			break;
		case OP_LSHIFT:
			s1->num.integer.i<<=s2->num.integer.i;
			break;
		case OP_RSHIFT:
			s1->num.integer.i>>=s2->num.integer.i;
			break;
		}
	}
	return 0;
}


/* integer ǂ̃`FbNĂȂ̂𒍈ӂ鎖 */
void conv_integer_to_float(struct stack_s * const s)
{
	s->type=TYPE_FLOAT;
	if(s->num.integer.sign == MINUS)
		s->num.f=-((myfloat)s->num.integer.i);
	else
		s->num.f=s->num.integer.i;
}




#ifndef RUPUTER
void code_to_string(struct lex_s *c, int count, char *str)
{
	static char op[][4]={
		"+",
		"-",
		"*",
		"/",
		"%",
		"**",
		"",
		"",
		"|",
		"&",
		"^",
		"<<",
		">>",
		"~",
		"#",
		"="
	};

	while(count--)
	{
		switch(c->type)
		{
		case TYPE_FLOAT:
#ifdef USE_LONG_DOUBLE
			{
				char *p;

				p=_ldtoa_r(c->data.num.f, 0, 24, NULL, NULL, NULL);
				strcpy(str, p);
			}
#else
			sprintf(str, "%e(%f)", c->data.num.f, c->data.num.f);
#endif
			break;
		case TYPE_INTEGER:
			integer_to_string((struct stack_s *)c, str);
			break;
		case TYPE_VARIABLE:
			strcat(str, c->data.id);
			break;
		case TYPE_UNDEF_VARIABLE:
	//			sprintf(str, "%s(undef) ", c->data.id);
			strcpy(str, c->data.id);
			strcat(str, "(undef) ");
			break;
		case TYPE_FUNCTION:
	//			sprintf(str, "%s() ", c->data.id);
			strcpy(str, c->data.id);
			strcat(str, "() ");
			break;
		case TYPE_PARAMETER:
			sprintf(str, "p%lu ", c->data.num.integer.i);
			break;
		case OP_COMMA:
	//			sprintf(str, ",");
			strcpy(str, ",");
			break;
		default:
			strcat(str, op[c->type]);
		}
		strcat(str, " ");
		str+=strlen(str);
		c++;
	}

}
#endif
#ifdef DEBUG_OUTPUT
void print_code(void)
{
	char str[1024];

	code_to_string(code, code_count, str);
	puts(str);
	fflush(stdout);
}

#endif
