#ifndef UNDER_CE
#include <stdio.h>
#include <errno.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "lex.h"
#include "variable.h"
#include "machine.h"
#include "../libmisc/misc.h"

#ifdef UNDER_CE
# if UNDER_CE <= 211
#include <windows.h>
int isalpha(int c);
int isspace(int c);
int isdigit(int c);
double strtod(const char *s, char **errorptr);
unsigned long strtoul(const char *s, char **errorptr, int base);
#endif
#endif



struct lex_s lex;


static void lex_ident(void);
static void lex_number(void);
static void lex_other(void);


static char *forward;	/* ̖͂ */
static char *lex_begin;	/* ̐擪 */


/*
 * ̓[`̏
 * ߂ĺÂƂ̎̐擪w|C^̃|C^B
 * ̎dĺAO[oȕϐo邾gȂƂB
 *
 */
void lex_init(char *str)
{
	forward=str;
	lex_begin=str;

}

char *lex_get_lex_begin(void)
{
	return lex_begin;
}

void lex_set_lex_begin(char * const p)
{
	lex_begin=p;
	forward=p;
}

void lex_next(void)
{

	while(isspace(*forward))
		forward++;

	lex_begin=forward;
	lex.aux=lex_begin;

	if(isdigit(*forward) || *forward == '.')
		lex_number();
	else if(isalpha(*forward) || *forward == '_')
		lex_ident();
	else
		lex_other();

}

static void lex_ident(void)
{
	char *p, *q;
	int len;

	p=forward;
	while(isalpha(*forward) || isdigit(*forward) || *forward == '_')
		forward++;

	len=forward-p;
	q=my_malloc(len+1);

	memcpy(q, p, len);
	q[len]=0;

	p=forward;
	while(isspace(*p))
		p++;

	if(*p == '(')
	{
		forward=p+1;
		lex.type=TYPE_FUNCTION;
	}
	else
	{
		struct variable_s *pv;

		pv=variable_get(q);
		if(pv == NULL)
		{
			lex.type=TYPE_UNDEF_VARIABLE;
		}
		else
		{
			lex.type=TYPE_VARIABLE;
		}
		lex.variable=pv;
	}

	lex.data.id=q;
}

static void lex_number(void)
{
	char *end;

	lex.type=TYPE_INTEGER;
	errno=0;
	lex.data.num.integer.i=strtoul(forward, &end, 0);
	if(*end == '.' || *end == 'e' || *end == 'E' || errno == ERANGE)
	{
		lex.type=TYPE_FLOAT;
		lex.data.num.f=strtod(forward, &end);
	}
	else
		lex.data.num.integer.sign=PLUS;

	forward=end;
}

static void lex_other(void)
{
	switch(*forward++)
	{
	case '+':	lex.type=TYPE_PLUS; 	break;
	case '-':	lex.type=TYPE_MINUS;	break;
	case '/':	lex.type=TYPE_DIV;		break;
	case '%':	lex.type=TYPE_MOD;		break;

	case '*':
		if(*forward == '*')
		{
			lex.type=TYPE_POWER;
			forward++;
		}
		else
			lex.type=TYPE_MUL;
		break;

	case ',':	lex.type=TYPE_COMMA;	break;
	case '=':	lex.type=TYPE_EQ;		break;

	case '(':	lex.type=TYPE_LPAR;		break;
	case ')':	lex.type=TYPE_RPAR;		break;

	case '|':	lex.type=TYPE_OR;		break;
	case '&':	lex.type=TYPE_AND;		break;
	case '^':	lex.type=TYPE_XOR;		break;
	case '~':	lex.type=TYPE_NOT;		break;
	case '>':
		if(*forward == '>')
		{
			lex.type=TYPE_RSHIFT;
			forward++;
		}
		else
			lex.type=TYPE_OTHER;
		break;
	case '<':
		if(*forward == '<')
		{
			lex.type=TYPE_LSHIFT;
			forward++;
		}
		else
			lex.type=TYPE_OTHER;
		break;

	case '\0':	lex.type=TYPE_EOS;		break;
	default:	lex.type=TYPE_OTHER;
	}
}


#ifdef UNDER_CE
#if UNDER_CE <= 211

int isalpha(int c);
int isspace(int c);
int isdigit(int c);
double strtod(const char *s, char **errorptr);
unsigned long strtoul(const char *s, char **errorptr, int base);

int isalpha(int c)
{
	if(c >= 'a' && c <= 'z'
	   || c >= 'A' && c >= 'Z')
		return 1;
	return 0;
}

int isspace(int c)
{
	if(c == ' ' && c == '\t')
		return 1;
	return 0;
}
int isdigit(int c)
{
	if(c >= '0' && c <= '9')
		return 1;
	return 0;
}


double strtod(const char *s, char **errorptr)
{
  double d;
  wchar_t *ws;
  wchar_t *werrorptr = NULL;
  int len = strlen(s) + 1;

  ws = malloc(len * sizeof(wchar_t));

  MultiByteToWideChar(CP_ACP, 0, s, -1, ws, len);

  d = wcstod(ws, &werrorptr);

  if(errorptr)
    {
      if(werrorptr)
	{
	  *errorptr = (char *) s + (werrorptr - ws);
	}
      else
	{
	  *errorptr = NULL;
	}
    }

  free(ws);

  return d;
}


unsigned long strtoul(const char *s, char **errorptr, int base)
{
  unsigned long l;
  wchar_t *ws;
  wchar_t *werrorptr = NULL;
  int len = strlen(s) + 1;

  // was malloc
  ws = _alloca(len * sizeof(wchar_t));

  MultiByteToWideChar(CP_ACP, 0, s, -1, ws, len);

  l = wcstoul(ws, &werrorptr, base);

  if(errorptr)
    {
      if(werrorptr)
	{
	  *errorptr = (char *) s + (werrorptr - ws);
	}
      else
	{
	  *errorptr = NULL;
	}
    }

  return l;
}


#endif // UNDER_CE
#endif // UNDER_CE
