


#ifndef UNDER_CE
#include <stdio.h>
#endif

#include <stdlib.h>

#include "function.h"
#include "parse.h"
#include "code.h"
#include "lex.h"
#include "stack.h"


static int parse_paren(int function);
static void check_rpar_error(void);
static void expr0(void);
static void expr1(void);
static void expr2(void);
static void expr3(void);
static void expr4(void);
static void expr5(void);
static void expr6(void);
static void expr7(void);
static void expr8(void);
static void expr9(void);
static void factor(void);

int error;

int def_function;
static int parameter_count;
static int in_function;

#define LV_CONST 	0
#define LV_VARIABLE	1
#define LV_FUNCTION	2

static int lvalue;

void expression(void)
{
	lvalue=LV_CONST;
	def_function=0;

	expr0();
}
/* comma */
static void expr0(void)
{
	expr1();
	while(lex.type == TYPE_COMMA)
	{
		if(error)
			return;

		parameter_count++;

		lvalue=LV_CONST;
		expr1();

		if(in_function == 0)	/* not function paren */
			add_code(OP_COMMA, NULL);
	}
}

static void expr1(void)
{
	expr2();
	if(lex.type == TYPE_EQ)
	{
		if(error)
			return;

		if(def_function == 1)
		{
			error=ER_SUBST_IN_FUNCTION;
			return;
		}

		if(lvalue == LV_CONST)
			error=ER_LVALUE_IS_CONST;
		else if(lvalue == LV_VARIABLE)
		{
			struct lex_s tmp;

			tmp=code[--code_count];
			expr1();
			add_code(OP_SUBST, &tmp);
		}
		else if(lvalue == LV_FUNCTION)
		{
			def_function=1;
//			add_code(OP_DEF_FUNCTION, NULL);
			expr1();
		}
	}
}

static void expr2(void)
{
	expr3();
	while(lex.type == TYPE_OR)
	{
		struct lex_s tmp;

		if(error)
			return;

		tmp=lex;
		expr3();
		lvalue=LV_CONST;
		add_code(OP_OR, &tmp);
	}
}

static void expr3(void)
{
	expr4();
	while(lex.type == TYPE_XOR)
	{
		struct lex_s tmp;

		if(error)
			return;

		tmp=lex;
		expr3();
		lvalue=LV_CONST;
		add_code(OP_XOR, &tmp);
	}
}

static void expr4(void)
{
	expr5();
	while(lex.type == TYPE_AND)
	{
		struct lex_s tmp;

		if(error)
			return;

		tmp=lex;
		expr5();
		lvalue=LV_CONST;
		add_code(OP_AND, &tmp);
	}
}

static void expr5(void)
{
	expr6();
	for(;;)
	{
		switch(lex.type)
		{
		case TYPE_LSHIFT:
		case TYPE_RSHIFT:
		{
			struct lex_s tmp;

			if(error)
				return;

			tmp=lex;

			expr6();
			lvalue=LV_CONST;
			add_code(tmp.type, &tmp);
			break;
		}
		default:
			return;
		}
	}
}

static void expr6(void)
{
	expr7();
	for(;;)
	{
		switch(lex.type)
		{
		case TYPE_PLUS:
		case TYPE_MINUS:
		{
			struct lex_s tmp;

			if(error)
				return;

			tmp=lex;
			expr7();
			lvalue=LV_CONST;
			add_code(tmp.type, &tmp);
			break;
		}
		default:
			return;
		}
	}
}
static void expr7(void)
{
	expr8();
	for(;;)
	{
		switch(lex.type)
		{
		case TYPE_MUL:
		case TYPE_DIV:
		case TYPE_MOD:
		{
			struct lex_s tmp;

			if(error)
				return;

			tmp=lex;
			expr8();
			lvalue=LV_CONST;
			add_code(tmp.type, &tmp);
			break;
		}
		default:
			return;
		}
	}
}
static void expr8(void)
{
	lex_next();
	switch(lex.type)
	{
	case TYPE_NOT:
	{
		struct lex_s tmp;

		tmp=lex;
		expr8();
		lvalue=LV_CONST;
		add_code(OP_NOT, &tmp);
		break;
	}
	case TYPE_MINUS:
	{
		struct lex_s tmp;

		tmp=lex;
		expr8();
		lvalue=LV_CONST;
		add_code(OP_INV, &tmp);
		break;
	}
	case TYPE_PLUS:
		do{
			lex_next();
		}while(lex.type == TYPE_PLUS);
/* break Ȃ̂ 킴 */
	default:
		expr9();
	}
}

static void expr9(void)
{
	factor();
	if(lex.type == TYPE_POWER)
	{
		struct lex_s tmp;

		if(error)
			return;

		tmp=lex;
		expr8();
		lvalue=LV_CONST;
		add_code(OP_POW, &tmp);
	}
}


static void factor(void)
{
	switch(lex.type)
	{
	case TYPE_LPAR:
		parse_paren(0);
		break;

	case TYPE_FLOAT:
	case TYPE_INTEGER:
		lvalue=LV_CONST;
		add_code(lex.type, &lex);
		lex_next();
		break;

	case TYPE_VARIABLE:
	case TYPE_UNDEF_VARIABLE:
		lvalue=LV_VARIABLE;
		add_code(lex.type, &lex);
		lex_next();
		break;

	case TYPE_FUNCTION:
	{
		struct lex_s tmp;
		int n, m;

		tmp=lex;

		n=parse_paren(1);
		if(error)
			break;
		m=check_parameter_number(tmp.data.id);
		if(m != 0 && m != n)
		{
			error=ER_PARAMETER;
			lex_set_lex_begin(tmp.aux);	/* erorr ȏꏊ֐̂Ƃ */
		}
		else
		{
			lvalue=LV_FUNCTION;
			add_code(TYPE_FUNCTION, &tmp);
		}

		break;
	}
	case TYPE_EOS:
		error=ER_EXPECT_NUMBER;
		break;
	case TYPE_OTHER:
	default:
		error=ER_UNACCEPTED_CHARACTER;
		break;
	}

	if(lex.type == TYPE_OTHER && error == 0)
		error=ER_UNACCEPTED_CHARACTER;
}
static int parse_paren(int function)
{
	int prev_parameter_count;
	int prev_in_function;
	int n;

	prev_in_function=in_function;
	in_function=function;

	prev_parameter_count=parameter_count;
	parameter_count=0;

	expr0();
	lvalue=LV_CONST;
	check_rpar_error();

	n=parameter_count+1;

	parameter_count=prev_parameter_count;
	in_function=prev_in_function;

	return n;
}

static void check_rpar_error(void)
{
	if(error)
		return;

	if(lex.type == TYPE_RPAR)
		lex_next();
	else
	{
		if(lex.type == TYPE_EOS)
			error=ER_NOT_RIGHT_PAREN;
		else
			error=ER_UNACCEPTED_CHARACTER;
	}
}

