#ifndef UNDER_CE
#include <stdio.h>
#endif

#include <stdlib.h>
#include <string.h>
#include "lex.h"
#include "../libmisc/misc.h"
#include "variable.h"
#include "code.h"
#include "user_func.h"
#include "parse.h"
#include "machine.h"

static int get_parameter_list(char * const name);
static char *set_parameter_list(void);
static struct user_function_s *make_user_function(char * const name,
												  struct lex_s * const func_body,
												  int func_code_count);
static void remake_user_function(struct user_function_s *p,
							   struct lex_s * const func_body,
							   int func_code_count);


static struct user_function_s root={NULL, 0, NULL, 0, NULL};

/* ֐̓o^
 *
 * code code_count B
 * +-+-+-+-+-+-+-+-+-+-+-      -+-+-+-+-+-+-+-+-+-+-+-+
 * |x|y|z|f|x|y|+| | | |  ....  | | | | | | | | | | | | 
 * +-+-+-+-+-+-+-+-+-+-+-      -+-+-+-+-+-+-+-+-+-+-+-+
 *    ֐̎
 *       ֐
 * f(x,y,z)=x+y ƁAȊB
 * n߁A́i֐o^̍ۂɎgjB́A
 * ɏ̂ɗBɁA֐B̌ォÂn܂B̂́A
 * ׂăRs[
 */

static char *parameter_list[16];
static int parameter_count;

char *func_def_user_function(void)
{
	char *aux;
	struct lex_s *func_code;
	struct lex_s *orig_code;
	struct user_function_s *p;
	int i;
	int func_code_count;

	parameter_count=0;

/* ֐̃R[hA̐
 * code_count = func_code_count+parameter_count+1
 * 𖞂B
 */
	aux=set_parameter_list();
	if(aux != NULL) /* error */
		return aux;

	orig_code=code+parameter_count;	/* orig_code ֐ */
	func_code_count=code_count-parameter_count-1;

	aux=orig_code->data.id;
	orig_code->data.id=NULL;	/* JȂ */

	func_code=my_malloc(func_code_count*sizeof(struct lex_s));
	orig_code++;
	memcpy(func_code, orig_code, func_code_count*sizeof(struct lex_s));

	for(i=0; i < func_code_count; i++)
	{
		switch(func_code[i].type)
		{
		case TYPE_FUNCTION:
			orig_code[i].data.id=NULL;
			break;
		case TYPE_UNDEF_VARIABLE:
		{
			int j;

			j=get_parameter_list(func_code[i].data.id);
			if(j < 0)
			{
				free(aux);
				free(func_code);
				error=ER_UNDEF_VARIABLE;
				return orig_code[i].aux;
			}

			func_code[i].type=TYPE_PARAMETER;
			func_code[i].data.num.integer.i=j;
			break;
		}
		case TYPE_VARIABLE:
		{
			int j;

			j=get_parameter_list(func_code[i].data.id);
			if(j > 0)
			{
/* ƈv */
				func_code[i].type=TYPE_PARAMETER;
				func_code[i].data.num.integer.i=j;
			}
			else
			{
/* ɂ͂ȂBŁA萔Ă */
				func_code[i].type=func_code[i].variable->type;
				func_code[i].data.num=func_code[i].variable->num;
			}

			break;
		}
		}
	}

	p=check_user_function(aux);
	if(p != NULL)
	{
/* aux ͈ȑOƓeȂ̂ŁAfree  */
		free(aux);
		remake_user_function(p, func_code, func_code_count);
	}
	else
	{

#ifdef RUPUTER
		rup_func_add_menu(aux);
#endif
		p=make_user_function(aux, func_code, func_code_count);

	}
	return NULL;
}

struct user_function_s *check_user_function(char * const name)
{
	struct user_function_s *p;

	p=root.next;

	for(; p != NULL; p=p->next)
		if(strcmp(p->name, name) == 0)
			return p;

	return NULL;
}


static void remake_user_function(struct user_function_s *p,
							   struct lex_s * const func_body,
							   int func_code_count)
{
	free(p->code);
	p->code=func_body;

	p->code_count=func_code_count;
	p->parameter_count=parameter_count;
}

static struct user_function_s *make_user_function(char * const name,
												  struct lex_s * const func_body,
												  int func_code_count)
{
	struct user_function_s *p;

	p=my_malloc(sizeof(struct user_function_s));

	p->name=name;

	p->code=func_body;

	p->code_count=func_code_count;
	p->parameter_count=parameter_count;

/* NɌq */
	p->next=root.next;
	root.next=p;

	return p;
}


static int get_parameter_list(char * const name)
{
	int i;

	for(i=0; i < parameter_count; i++)
		if(strcmp(parameter_list[i], name) == 0)
			return i;
	return -1;
}

static char *set_parameter_list(void)
{
	int i;

	for(i=0; i < code_count; i++)
	{
		switch(code[i].type)
		{
		case TYPE_FUNCTION:
			parameter_count=i;
			return NULL;
		case TYPE_VARIABLE:
		case TYPE_UNDEF_VARIABLE:
			parameter_list[parameter_count++]=code[i].data.id;
			break;
		default:
			error=ER_PARAMETER_IS_CONST;
			return code[i].aux;
		}
	}
	return NULL;
}

#ifdef ENABLE_FUNCTION_TO_STRING
void function_to_string(char *str)
{
	struct user_function_s *p;

	*str=0;

	p=root.next;
	for(; p != NULL; p=p->next)
	{
//		sprintf(str, "%s()=", p->name);
		strcpy(str, p->name);
		strcat(str, "()=");

		code_to_string(p->code, p->code_count, str+strlen(str));

#if defined(WINDOWS) || defined(UNICODE)
		strcat(str, "\x0d\x0a");
#else
		strcat(str, "\n");
#endif
		str+=strlen(str);
	}
}
#endif

